import logging

import click
from rich import get_console
from rich.logging import RichHandler


MAX_LOG_LEVEL = 3

OPTION_KWARGS_VERBOSE = dict(
    count=True,
    type=click.IntRange(0, MAX_LOG_LEVEL),
    default=0,
    help='Verbosity level; Accepts integer value e.g.: "--verbose 2" or can be count e.g.: "-vv" ',
    show_default=True,
)


def setup_logging(*, verbosity: int):
    log_format = '%(message)s'
    if verbosity == 0:
        level = logging.ERROR
    elif verbosity == 1:
        level = logging.WARNING
    elif verbosity == 2:
        level = logging.INFO
    else:
        level = logging.DEBUG
        log_format = '(%(name)s) %(message)s'

    console = get_console()
    console.print(f'(Set log level {verbosity}: {logging.getLevelName(level)})', justify='right')
    logging.basicConfig(
        level=level,
        format=log_format,
        datefmt='[%x %X.%f]',
        handlers=[RichHandler(console=console, omit_repeated_times=False)],
        force=True,
    )



from pathlib import Path


def replace_prefix(text: str, *, prefix: str, new_prefix: str) -> str:
    """
    >>> replace_prefix('foobar', prefix='foo', new_prefix='FOO')
    'FOObar'
    >>> replace_prefix('foobar', prefix='no', new_prefix='yes')
    'foobar'
    """
    if text.startswith(prefix):
        return f'{new_prefix}{text.removeprefix(prefix)}'
    return text


def replace_dict_values_prefix(data: dict, *, prefix: str, new_prefix: str) -> None:
    """
    Replace prefixes in dict values, recusive and in-place. Works with Path() values, too.

    Usecase: You have a dict structure with many temp file path
    and would like to replace all random temp path prefixes ;)

    >>> data = {'foo': '123FOO', 'bar': {'baz': '123BAR'}}
    >>> replace_dict_values_prefix(data, prefix='123', new_prefix='xxx')
    >>> data
    {'foo': 'xxxFOO', 'bar': {'baz': 'xxxBAR'}}
    """
    for key, value in data.items():
        if isinstance(value, str) and value.startswith(prefix):
            data[key] = replace_prefix(value, prefix=prefix, new_prefix=new_prefix)
        elif isinstance(value, Path) and str(value).startswith(prefix):
            data[key] = Path(replace_prefix(str(value), prefix=prefix, new_prefix=new_prefix))
        elif isinstance(value, dict):
            replace_dict_values_prefix(value, prefix=prefix, new_prefix=new_prefix)



import dataclasses
from pathlib import Path

from bx_py_utils.environ import OverrideEnviron

from ha_services.cli_tools.test_utils.environment_fixtures import MockCurrentWorkDir
from ha_services.toml_settings.api import TomlSettings
from ha_services.toml_settings.serialize import dataclass2toml_str


def replace_dataclass_values(instance, *, data: dict):
    """
    Replace dataclass values in-place with values from given dict.
    """
    assert dataclasses.is_dataclass(instance), f'{instance=}'

    for key, value in data.items():
        if isinstance(value, dict):
            sub_dataclass = getattr(instance, key)
            replace_dataclass_values(sub_dataclass, data=value)
        else:
            assert hasattr(instance, key), f'Attribute "{key}" missing on: {instance=}'
            setattr(instance, key, value)


def replace_path_values(instance):
    assert dataclasses.is_dataclass(instance), f'{instance=}'

    data = dataclasses.asdict(instance)
    for key, value in data.items():
        if isinstance(value, dict):
            sub_dataclass = getattr(instance, key)
            replace_path_values(sub_dataclass)
        elif isinstance(value, Path):
            setattr(instance, key, str(value))


class MockTomlSettings(MockCurrentWorkDir):
    """
    Mock TomlSettings with overwrites of settings dataclass instance.
    Will create a "default" settings.toml in a temporary directory.
    """

    def __init__(
        self,
        *,
        SettingsDataclass,
        settings_overwrites: dict,  # Change values in SettingsDataclass instance
        dir_name='mocked_dir_name',
        file_name='mocked_file_name',
        **temp_kwargs,
    ):
        super().__init__(**temp_kwargs)

        assert dataclasses.is_dataclass(SettingsDataclass), f'{SettingsDataclass=}'
        self.SettingsDataclass = SettingsDataclass
        self.settings_overwrites = settings_overwrites
        self.dir_name = dir_name
        self.file_name = file_name

        self.overwrite_env = None
        self.toml_settings = None
        self.settings_dataclass = None

    def __enter__(self) -> 'MockTomlSettings':
        super().__enter__()

        self.overwrite_env = OverrideEnviron(HOME=str(self.temp_path))
        self.overwrite_env.__enter__()

        self.config_path = self.temp_path / '.config'
        self.config_path.mkdir()

        self.settings_dataclass = self.SettingsDataclass()
        replace_dataclass_values(instance=self.settings_dataclass, data=self.settings_overwrites)

        self.toml_settings = TomlSettings(
            dir_name=self.dir_name,
            file_name=self.file_name,
            settings_dataclass=self.settings_dataclass,
        )

        doc_str = dataclass2toml_str(instance=self.settings_dataclass)
        self.toml_settings.file_path.write_text(doc_str, encoding='UTF-8')

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)
        self.overwrite_env.__exit__(exc_type, exc_val, exc_tb)
        if exc_type:
            return False



import getpass
import os
import shlex
import sys
import tempfile
from pathlib import Path

from bx_py_utils.environ import OverrideEnviron


class AsSudoCallOverrideEnviron(OverrideEnviron):
    """
    Manipulate the environment variables so it looks like a "sudo" call.
    Compare with: "sudo env | sort" ;)
    """

    def __init__(self, **overrides):
        overrides = {
            'SUDO_COMMAND': shlex.join(sys.argv),
            'SUDO_UID': str(os.getuid()),
            'SUDO_GID': str(os.getgid()),
            'SUDO_USER': getpass.getuser(),
            'LOGNAME': 'root',
            'HOME': '/root',
            **overrides,
        }
        super().__init__(**overrides)


class MockCurrentWorkDir(tempfile.TemporaryDirectory):
    """
    Context Manager to move the "CWD" to a temp directory.
    """

    def __init__(self, **kwargs):
        self.old_cwd = Path().cwd()
        super().__init__(**kwargs)

    def __enter__(self):
        temp_dir = super().__enter__()
        os.chdir(temp_dir)
        self.temp_path = Path(temp_dir)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)
        os.chdir(self.old_cwd)
        if exc_type:
            return False



from pathlib import Path
from unittest import TestCase

from ha_services.cli_tools.dict_utils import replace_dict_values_prefix


class DictUtilsTestCase(TestCase):
    def test_replace_dict_values_prefix(self):
        data = {'foo': '123FOO', 'bar': {'baz': Path('123BAR')}}
        replace_dict_values_prefix(data, prefix='123', new_prefix='xxx')
        self.assertEqual(data, {'foo': 'xxxFOO', 'bar': {'baz': Path('xxxBAR')}})



import shutil
from unittest import mock

from bx_py_utils.environ import OverrideEnviron
from bx_py_utils.test_utils.context_managers import MassContextManager

from ha_services.cli_tools.test_utils.environment_fixtures import MockCurrentWorkDir
from ha_services.systemd import defaults
from ha_services.systemd.data_classes import BaseSystemdServiceInfo


class MockedSys:
    executable = '/mocked/.venv/bin/python3'


class MockSystemdServiceInfo(MassContextManager):
    """
    Set all values in ha_services/systemd/defaults.py to static ones,
    independent of current user/environment.

    So that creating a SystemdServiceInfo() instance will result in a well known state.

    Note: The work_dir has still a random suffix!
    """

    def __init__(self, *, prefix: str, SystemdServiceInfoClass):
        self.mocked_cwd = MockCurrentWorkDir(prefix=prefix)
        assert issubclass(SystemdServiceInfoClass, BaseSystemdServiceInfo), f'{SystemdServiceInfoClass=}'
        self.SystemdServiceInfoClass = SystemdServiceInfoClass

        self.mocks = (
            self.mocked_cwd,
            OverrideEnviron(SUDO_USER='MockedUserName'),
            mock.patch.object(defaults, 'sys', MockedSys()),
        )

        self.temp_path = None
        self.systemd_info = None

    def __enter__(self) -> 'MockSystemdServiceInfo':
        super().__enter__()
        self.temp_path = self.mocked_cwd.temp_path
        mocked_systemd_base_path = self.temp_path / 'etc-systemd-system'
        mocked_systemd_base_path.mkdir()
        self.systemd_info: BaseSystemdServiceInfo = self.SystemdServiceInfoClass(
            systemd_base_path=mocked_systemd_base_path  # noqa
        )

        # Move the source template into temp directory:
        src_template_path = self.systemd_info.template_path
        dst_template_path = self.temp_path / src_template_path.name
        shutil.copyfile(src_template_path, dst_template_path)
        self.systemd_info.template_path = dst_template_path

        return self



import os
import subprocess
import sys
from unittest.mock import patch

import rich
import rich_click
from bx_py_utils.environ import OverrideEnviron
from bx_py_utils.path import assert_is_file
from bx_py_utils.test_utils.context_managers import MassContextManager
from click._compat import strip_ansi as strip_ansi_codes
from manageprojects.utilities.subprocess_utils import verbose_check_output
from rich import get_console
from rich.console import Console, get_windows_console_features

from ha_services.cli_tools.test_utils.assertion import assert_in


BASE_WIDTH = 120

FIX_TERM_DICT = dict(
    PYTHONUNBUFFERED='1',
    COLUMNS=str(BASE_WIDTH),
    TERM='dump',
    NO_COLOR='1',
    COLORTERM=None,
    FORCE_COLOR=None,
)


def get_fixed_env_copy(width: int = BASE_WIDTH, exclude_none=False) -> dict:
    env_dict = FIX_TERM_DICT.copy()
    env_dict['COLUMNS'] = str(width)
    if exclude_none:
        env_dict = {key: value for key, value in env_dict.items() if value is not None}
    return env_dict


class NoColorTermEnviron(MassContextManager):
    """
    Context manager that set all environment variables to deactivate terminal colors and fix the terminal width
    """

    def __init__(self, width: int = BASE_WIDTH):
        self.mocks = [
            OverrideEnviron(
                **get_fixed_env_copy(width),
            ),
        ]


class NoColorEnvRich(NoColorTermEnviron):
    """
    Context manager that patch "rich" console to a no-color with a defined width.
    """

    def __init__(self, width: int = BASE_WIDTH):
        super().__init__(width)
        self.mocks += [
            patch.object(rich, '_console', Console(width=width, no_color=True)),
            patch.object(Console, '_environ', get_fixed_env_copy(width, exclude_none=True)),
        ]


class NoColorEnvRichClick(NoColorEnvRich):
    """
    Context manager for rich-click to deactivate terminal colors and fix the terminal width
    """

    def __init__(self, width: int = BASE_WIDTH):
        super().__init__(width)
        self.mocks += [
            patch.object(rich_click.rich_click, 'MAX_WIDTH', width),
            patch.object(rich_click.rich_click, 'FORCE_TERMINAL', False),
        ]


class NoColorRichClickCli(NoColorEnvRichClick):
    """
    Context manager to get the output of a rich-click CLI command.
    To re-evaluate module-level code, call the CLI via a subprocess.
    """

    def invoke(self, *, cli_bin, args, strip_line_prefix: str = '', exit_on_error=True) -> str:
        assert_is_file(cli_bin)

        stdout = verbose_check_output(
            cli_bin,
            *args,
            cwd=cli_bin.parent,
            exit_on_error=exit_on_error,
        )

        stdout = strip_ansi_codes(stdout)  # FIXME: Needed on github CI, why?!?

        if strip_line_prefix:
            # Skip header lines:
            lines = stdout.splitlines()
            found = False
            for pos, line in enumerate(lines):
                if line.lstrip().startswith(strip_line_prefix):
                    stdout = '\n'.join(lines[pos:])
                    found = True
                    break

            assert found is True, f'Line that starts with {strip_line_prefix=} not found in: {stdout!r}'

            stdout = '\n'.join(line.rstrip() for line in stdout.splitlines())

        return stdout


def assert_no_color_env(*, width: int) -> None:
    assert os.environ['COLUMNS'] == str(width), f'{os.environ["COLUMNS"]=} is not "{width=}"'
    assert os.environ['TERM'] == 'dump', f'{os.environ["TERM"]=} is not "dump"'
    assert os.environ['NO_COLOR'] == '1', f'{os.environ["NO_COLOR"]=} is not "1"'


def assert_subprocess_rich_diagnose_no_color(*, width: int) -> None:
    output = subprocess.check_output(
        [sys.executable, '-m', 'rich.diagnose'],
        text=True,
        env=os.environ.copy(),  # It's needed to pass the current environment!
    )
    assert_in(
        output,
        parts=(
            f'width={width}',
            'color_system = None',
            'no_color = True',
            'highlight=None,',
            'style = None',
            'truecolor = False',
            'vt = False',
        ),
    )


def assert_rich_no_color(*, width: int) -> None:
    features = get_windows_console_features()
    assert features.truecolor is False, f'{features=}'

    global_console = get_console()
    assert global_console._width == width, f'{global_console._width=} is not {width=}'
    assert global_console.no_color is True

    new_console = Console()
    assert new_console._width == width, f'{new_console._width=} is not {width=}'
    assert new_console.no_color is True


def assert_rich_click_no_color(*, width: int) -> None:
    assert rich_click.rich_click.MAX_WIDTH == width, f'{rich_click.rich_click.MAX_WIDTH=} is not "{width}"'
    assert rich_click.rich_click.FORCE_TERMINAL is False, f'{rich_click.rich_click.FORCE_TERMINAL=} is not False'



from pathlib import Path
from unittest import TestCase

from bx_py_utils.path import assert_is_dir, assert_is_file

from ha_services.cli.cli_app import SETTINGS_DIR_NAME, SETTINGS_FILE_NAME
from ha_services.example import DemoSettings
from ha_services.toml_settings.api import TomlSettings
from ha_services.toml_settings.test_utils.data_class_utils import (
    MockTomlSettings,
    replace_dataclass_values,
    replace_path_values,
)
from ha_services.toml_settings.tests.fixtures import ComplexExample, PathExample2


class TestUtilsTestCase(TestCase):
    def test_replace_dataclass_values(self):
        instance = ComplexExample()

        # Check before:
        self.assertEqual(instance.foo, 'bar')
        self.assertEqual(instance.sub_class_one.number, 123)
        self.assertEqual(instance.sub_class_two.something, 0.5)

        new_data = {
            'foo': 'NEW',
            'sub_class_two': {
                'something': 456,
            },
        }
        replace_dataclass_values(instance, data=new_data)

        # Check after:
        self.assertEqual(instance.foo, 'NEW')
        self.assertEqual(instance.sub_class_one.number, 123)  # unchanged?
        self.assertEqual(instance.sub_class_two.something, 456)

    def test_replace_path_values(self):
        instance = PathExample2()

        # Check before:
        self.assertEqual(instance.path, Path('/foo/baz'))
        self.assertEqual(instance.sub_path.path, Path('/foo/bar'))

        replace_path_values(instance)

        # Check after:
        self.assertEqual(instance.path, '/foo/baz')
        self.assertEqual(instance.sub_path.path, '/foo/bar')

    def test_toml_settings_mock(self):
        settings_overwrites = dict(
            systemd=dict(
                template_context=dict(
                    user='MockedUserName',
                    group='MockedUserName',
                )
            ),
        )
        with MockTomlSettings(
            SettingsDataclass=DemoSettings,
            settings_overwrites=settings_overwrites,
            dir_name=SETTINGS_DIR_NAME,
            file_name=SETTINGS_FILE_NAME,
            prefix='test_mock',
        ) as cm:
            self.assertIsInstance(cm, MockTomlSettings)
            self.assertIsInstance(cm.temp_path, Path)
            assert_is_dir(cm.temp_path)

            toml_settings = cm.toml_settings
            self.assertIsInstance(toml_settings, TomlSettings)
            assert_is_file(toml_settings.file_path)

            settings = toml_settings.settings_dataclass
            self.assertIsInstance(settings, DemoSettings)

            # Check some sample values:
            self.assertEqual(settings.systemd.service_slug, 'haservices_demo')
            self.assertEqual(settings.systemd.template_context.verbose_service_name, 'HaServices Demo')
            self.assertEqual(settings.systemd.template_context.user, 'MockedUserName')
            self.assertEqual(settings.systemd.template_context.group, 'MockedUserName')
            self.assertEqual(settings.systemd.template_context.work_dir, cm.temp_path)
            self.assertEqual(settings.app.device_name, 'ha-services-demo')

        self.assertFalse(cm.temp_path.exists())



import sys
from pathlib import Path

from manageprojects.utilities.subprocess_utils import verbose_check_call
from rich import print


def clean_coverage_files():
    """
    Old / obsolete coverage files may corrupt the test run,
    so better try to remove them
    """
    for file_path in Path().cwd().glob('.coverage*'):
        try:
            file_path.unlink()
        except OSError as err:
            print(f'(Error remove {file_path}: {err})')
        else:
            print(f'(remove {file_path}, ok)')


def _run_unittest_cli(extra_env=None, verbose=True, exit_after_run=True):
    """
    Call the origin unittest CLI and pass all args to it.
    """
    clean_coverage_files()

    if extra_env is None:
        extra_env = dict()

    extra_env.update(
        dict(
            PYTHONUNBUFFERED='1',
            PYTHONWARNINGS='always',
        )
    )

    args = sys.argv[2:]
    if not args:
        if verbose:
            args = ('--verbose', '--locals', '--buffer')
        else:
            args = ('--locals', '--buffer')

    verbose_check_call(
        sys.executable,
        '-m',
        'unittest',
        *args,
        timeout=15 * 60,
        extra_env=extra_env,
    )
    if exit_after_run:
        sys.exit(0)


def _run_tox():
    """
    Call tox and pass all command arguments to it
    """
    clean_coverage_files()
    verbose_check_call(sys.executable, '-m', 'tox', *sys.argv[2:])
    sys.exit(0)



import os
import unittest.util

from ha_services.cli_tools.verbosity import MAX_LOG_LEVEL, setup_logging


# Hacky way to expand the failed test output:
unittest.util._MAX_LENGTH = os.environ.get('UNITTEST_MAX_LENGTH', 300)


# Display DEBUG logs in tests:
setup_logging(verbosity=MAX_LOG_LEVEL)



from pathlib import Path

from bx_py_utils.auto_doc import assert_readme_block
from bx_py_utils.path import assert_is_file


class AssertCliHelpInReadme:
    README_FILE_NAME = 'README.md'

    def __init__(self, *, base_path: Path, cli_epilog: str):
        self.readme_path = base_path / self.README_FILE_NAME
        assert_is_file(self.readme_path)

        self.cli_epilog = cli_epilog

    def assert_block(self, text_block: str, marker: str):
        if self.cli_epilog:
            text_block = text_block.replace(self.cli_epilog, '')

        text_block = f'```\n{text_block.strip()}\n```'
        assert_readme_block(
            readme_path=self.readme_path,
            text_block=text_block,
            start_marker_line=f'[comment]: <> (✂✂✂ auto generated {marker} start ✂✂✂)',
            end_marker_line=f'[comment]: <> (✂✂✂ auto generated {marker} end ✂✂✂)',
        )



from ha_services.cli_tools.test_utils.rich_test_utils import BASE_WIDTH, NoColorRichClickCli
from ha_services.toml_settings.test_utils.data_class_utils import MockTomlSettings


class TomlSettingsCliMock(NoColorRichClickCli):
    """
    IMPORTANT: We must ensure that no local user settings added to the help text
    So we can't directly invoke_click() here, because user settings are read and
    used on module level!
    So we must use subprocess and use a default settings file!
    """

    def __init__(
        self,
        *,
        SettingsDataclass,
        settings_overwrites: dict,  # Change values in SettingsDataclass instance
        dir_name='mocked_dir_name',
        file_name='mocked_file_name',
        prefix='test_mock',  # Temp dir prefix
        width: int = BASE_WIDTH,  # Terminal max width
    ):
        super().__init__(width)
        self.mocks += [
            MockTomlSettings(
                SettingsDataclass=SettingsDataclass,
                settings_overwrites=settings_overwrites,
                dir_name=dir_name,
                file_name=file_name,
                prefix=prefix,
            ),
        ]



import dataclasses
from pathlib import Path
from unittest import TestCase

from bx_py_utils.path import assert_is_dir, assert_is_file
from bx_py_utils.test_utils.snapshot import assert_text_snapshot

from ha_services.cli_tools.dict_utils import replace_dict_values_prefix
from ha_services.systemd.data_classes import BaseSystemdServiceInfo, BaseSystemdServiceTemplateContext
from ha_services.systemd.test_utils.mock_systemd_info import MockSystemdServiceInfo
from ha_services.toml_settings.serialize import dataclass2toml_str
from ha_services.toml_settings.test_utils.data_class_utils import replace_dataclass_values


@dataclasses.dataclass
class SystemdServiceTemplateContext(BaseSystemdServiceTemplateContext):
    """Systemd template context - Test dataclass"""

    verbose_service_name: str = 'Service Info Mock Test'


@dataclasses.dataclass
class TestSystemdServiceInfo(BaseSystemdServiceInfo):
    """Systemd service info - Test dataclass"""

    template_context: SystemdServiceTemplateContext = dataclasses.field(default_factory=SystemdServiceTemplateContext)


class MocksTest(TestCase):
    def test_mock(self):
        with MockSystemdServiceInfo(prefix='test_mock', SystemdServiceInfoClass=TestSystemdServiceInfo) as cm:
            self.assertIsInstance(cm, MockSystemdServiceInfo)

            temp_path = cm.temp_path
            self.assertIsInstance(temp_path, Path)
            self.assertEqual(temp_path, Path.cwd())
            assert_is_dir(temp_path)

            systemd_info = cm.systemd_info
            self.assertIsInstance(systemd_info, TestSystemdServiceInfo)

            # Check some samples:
            self.assertEqual(systemd_info.template_context.verbose_service_name, 'Service Info Mock Test')
            self.assertEqual(systemd_info.template_context.user, 'MockedUserName')
            self.assertEqual(systemd_info.template_context.group, 'MockedUserName')
            self.assertEqual(
                systemd_info.template_context.exec_start, '/mocked/.venv/bin/python3 -m ha_services_app publish-loop'
            )
            self.assertEqual(systemd_info.service_slug, 'service_info_mock_test')
            assert_is_file(systemd_info.template_path)

            data = dataclasses.asdict(systemd_info)

            # Replace the random temp path:
            replace_dict_values_prefix(data, prefix=str(temp_path), new_prefix='/mocked-temp-path/')

            # Check some samples, again:
            self.assertEqual(data['template_path'], Path('/mocked-temp-path/service_template.txt'))
            self.assertEqual(
                data['service_file_path'], Path('/mocked-temp-path/etc-systemd-system/service_info_mock_test.service')
            )

            # Replace the random temp path in dataclass instance:
            replace_dataclass_values(systemd_info, data=data)

            doc_str = dataclass2toml_str(systemd_info)
            assert_text_snapshot(got=doc_str, extension='.toml')

        self.assertFalse(temp_path.exists())



from unittest import TestCase

from ha_services import __version__
from ha_services.cli.cli_app import PACKAGE_ROOT
from ha_services.cli_tools.test_utils.assertion import assert_in, assert_startswith
from ha_services.cli_tools.test_utils.rich_test_utils import (
    NoColorEnvRich,
    NoColorEnvRichClick,
    NoColorRichClickCli,
    NoColorTermEnviron,
    assert_no_color_env,
    assert_rich_click_no_color,
    assert_rich_no_color,
    assert_subprocess_rich_diagnose_no_color,
)


class MockRichTestCase(TestCase):
    def test_NoColorTermEnviron(self):
        with NoColorTermEnviron(width=100):
            assert_no_color_env(width=100)
            assert_subprocess_rich_diagnose_no_color(width=100)

    def test_NoColorEnvRich(self):
        with NoColorEnvRich(width=100):
            assert_no_color_env(width=100)
            assert_subprocess_rich_diagnose_no_color(width=100)

            assert_rich_no_color(width=100)

    def test_NoColorEnvRichClick(self):
        with NoColorEnvRichClick(width=100):
            assert_no_color_env(width=100)
            assert_subprocess_rich_diagnose_no_color(width=100)
            assert_rich_no_color(width=100)

            assert_rich_click_no_color(width=100)

    def test_NoColorRichClickCli(self):
        with NoColorRichClickCli() as cm:
            # Without "strip_line_prefix":
            stdout = cm.invoke(cli_bin=PACKAGE_ROOT / 'dev-cli.py', args=('--help',))
            assert_in(
                stdout,
                parts=(
                    '.venv/bin/ha_services_dev --help',
                    f'ha_services v{__version__}',
                    'Usage: ./dev-cli.py',
                    'Show this message and exit.',
                ),
            )
            assert_startswith(stdout, f'\n+ {PACKAGE_ROOT}')

            # Remove prefix lines:
            stdout = cm.invoke(cli_bin=PACKAGE_ROOT / 'dev-cli.py', args=('--help',), strip_line_prefix='Usage: ')
            assert_in(
                stdout,
                parts=(
                    'Usage: ./dev-cli.py',
                    'Show this message and exit.',
                ),
            )
            assert_startswith(stdout, ' Usage: ./dev-cli.py ')
